
/*
 * include important headers 
 */
#include <net-snmp/net-snmp-config.h>
#if HAVE_STDLIB_H
	#include <stdlib.h>
#endif
#if HAVE_STRING_H
	#include <string.h>
#else
	#include <strings.h>
#endif

/*
 * needed by util_funcs.h 
 */
#if TIME_WITH_SYS_TIME
	#ifdef WIN32
		#include <sys/timeb.h>
	#else
		#include <sys/time.h>
	#endif
	#include <time.h>
#else
	#if HAVE_SYS_TIME_H
		#include <sys/time.h>
	#else
		#include <time.h>
	#endif
#endif

#if HAVE_WINSOCK_H
	#include <winsock.h>
#endif
#if HAVE_NETINET_IN_H
	#include <netinet/in.h>
#endif

#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>



/*
 * header_generic() comes from here 
 */
#include "util_funcs.h"

/*
 * include our .h file 
 */
#include <iostream>
#include <vector>
#include "arecamib.h"
#include "arclib.h"

#if defined(__linux__)
	#include "linux_comm.h"
	#include "linux_ioctl.h"
	#include "linux_pass_ioctl.h"
#endif	


#if defined(_WIN32)
	#include "win_comm.h"
	#include "win_pass_ioctl.h"
	#include "win_scsi.h"
#endif


#include "oid-sata.h"
#include "oid-sas.h"


class CArclibSNMP : public CArclib
{
private:
	oid idx;
public:
	void setIdx(oid x)
	{
		idx = x;
	};

	oid getIdx()
	{
		return idx;
	}  
};

#define	ARECASTRING		    1
#define ARECAINTEGER		21
/*
 *  Certain objects can be set via configuration file directives.
 *  These variables hold the values for such objects, as they need to
 *   be accessible to both the config handlers, and the callback routine.
 */
#define ARECA_COMMUNITY_LEN	        256
#define MAX_CONTROLLER_SUPPORTED    8

#define OIDBUF	64

CArclibSNMP ar[MAX_CONTROLLER_SUPPORTED];
static int  arrRaidCard[MAX_CONTROLLER_SUPPORTED];
static int  arrSASRaidCard[MAX_CONTROLLER_SUPPORTED];
static int  gTotalCtrl = 0;
static int  gTotalSASCtrl = 0;
static int  gCurrentCtrl = 0;      
//static char gOidBuf[1024];  
static char string_ret[256];	/* for ARECASTRING   */
static long long_ret;			/* for everything else */             

/*
 * This array defines the OID of the top of the mib tree that we're
 *  registering underneath.
 * Note that this needs to be the correct size for the OID being 
 *  registered, so that the length of the OID can be calculated.
 *  The format given here is the simplest way to achieve this.
 */
oid             areca_variables_oid[] = { 1, 3, 6, 1, 4, 1, 18928, 1, 1};


#ifdef __cplusplus
extern "C" {
#endif

using namespace std;


/*
 * Forward declarations for the config handlers 
 */
void            areca_parse_config_rocommunity(const char *token,
											   char *cptr);
void            areca_parse_config_rwcommunity(const char *token,
											   char *cptr);
void            areca_parse_config_evtlevel(const char *token, char *cptr);

void            EventMonitor(unsigned int clientreg, void *clientarg);



// the default community and event notification level
char gCommunity[ARECA_COMMUNITY_LEN] = "public";
int  gEvt_level = 0;

/*********************
*
*  Initialisation & common implementation functions
*
*********************/

char *arctime(const struct tm *timeptr)
{
	static char result[26];

	memset(result, 0, sizeof(result));

	sprintf(result, "%4d-%02d-%02d %02d:%02d:%02d", 
			timeptr->tm_year+1900,
			timeptr->tm_mon,
			timeptr->tm_mday,
			timeptr->tm_hour,
			timeptr->tm_min,
			timeptr->tm_sec);

	return result;
}

#if defined(_WIN32)

int DiscoveryDevice()
{
	WinPassIoctlInterface *iface = new WinPassIoctlInterface();
	for ( int i = 0; i < 16 ; i++ )
	{
		if ( iface->init(i) )
		{
			ar[gTotalCtrl++].ArcInitSession(iface);
			if ( gTotalCtrl >= MAX_CONTROLLER_SUPPORTED )
			{
				break;
			}

			iface = new WinPassIoctlInterface();                       
		}
	}

	return gTotalCtrl;

}

#elif defined(__linux__)

/************************************************************************************
	Routine Name:	DiscoveryCardDevice
	Function:		Discovery the controller(s) connected to the PCI, PCI-X, PCI-E bus
					through pass ioctl interface
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			Added by hank 2006-03-07
************************************************************************************/
int DiscoveryCardDeviceThroughPassIoctl()
{
	#define MAX_SCAN_SG 256	
	int sg = 0;
	char stat;

	do
	{
		LinuxPassIoctlInterface  *iface = new LinuxPassIoctlInterface();

		stat = (int)iface->init(sg);
		if ( stat )
		{
			ar[gTotalCtrl].ArcInitSession(iface);
			gTotalCtrl++;
			//break;
		}
		else
		{
			delete iface;
		}
	}while ( ++sg <= MAX_SCAN_SG && gTotalCtrl < MAX_CONTROLLER_SUPPORTED );

	return gTotalCtrl;

}

int DiscoveryDevice()
{
	const char  areca_driver_name[] = "arcmsr";
	int         major_num = 0;
	int         minor_num = 0;

	char        major_num_str[4];
	char        driver_str[7];
	char        dev_name[32];
	struct stat stat_of_dev;
	int         controller_index = 0;
	mode_t      mode = 0666;


	FILE *devices_file = NULL;

	devices_file = fopen("/proc/devices", "r");
	if ( !devices_file )
	{
		return 0;
	}


	// get areca's major#
	while ( fscanf(devices_file, "%3s%32s", major_num_str, driver_str) != EOF )
	{
		major_num_str[4] = 0;
		dev_name[7] = 0;

		// find areca's major#
		if ( !memcmp(areca_driver_name, driver_str, 6) )
		{
			major_num = atoi(major_num_str);            
			break;
		}
	}

	fclose(devices_file);

	if ( !major_num )
	{
		return 0;
	}

	mode |= S_IFCHR;

	for ( minor_num = 0; minor_num < MAX_CONTROLLER_SUPPORTED; minor_num++ )
	{
		sprintf(dev_name, "/dev/arcmsr%d", minor_num);
		if ( !stat(dev_name, &stat_of_dev) )
		{
			// device name found, check to see if the major and minor were correct.
			if ( major_num != (int)major(stat_of_dev.st_rdev) || 
				 minor_num != (int)minor(stat_of_dev.st_rdev) )
			{
				// major and minor number are not match, delete the old device name
				unlink(dev_name);
			}
			else
			{
				continue;
			}
		}

		if ( mknod(dev_name, mode, makedev(major_num, minor_num)) )
		{
			return 0;
		}
	}

	// setup areca controller(s)
	for ( controller_index = 0; controller_index < MAX_CONTROLLER_SUPPORTED; controller_index++ )
	{
		LinuxIoctlInterface *iface = new LinuxIoctlInterface();
		if ( iface )
		{
			if ( iface->init(controller_index) )
			{
				// call ArcInitSession(...) member function first before call any others member functions
				ar[controller_index].ArcInitSession(iface);

				// save the total vaild device we found
				gTotalCtrl++;
			}
			else
			{
				delete iface;
				iface = NULL;
			} // if
		}// if
		else
		{
			return 0;
		}
	}// for

	return gTotalCtrl;
}

#endif

#if defined(__FreeBSD__)

int DiscoveryDevice()
{
	char buf[64];
	char rmstr[128];
	mode_t mode = 0666;
	dev_t dev;
	int     nNumofPciDevs = 0;

	mode |= S_IFCHR;


	// remove all exist device node first
	for (int i = 0; i < MAX_CONTROLLER_SUPPORTED; i++)
	{
		memset(buf, 0, sizeof(buf));
		sprintf(buf,"/dev/arcmsr%d", i);
		unlink(buf);
	}


	LinuxIoctlInterface *out = new LinuxIoctlInterface();

	for (int mymajor = 255; mymajor > 161; mymajor--)
	{
		for (int myminor = 0; myminor < MAX_CONTROLLER_SUPPORTED; myminor++)
		{
			memset(buf, 0, sizeof(buf));
			sprintf(buf,"/dev/arcmsr%d", gTotalCtrl);
			dev = makedev(mymajor, myminor);
			mknod(buf, mode, dev);
			if (out->init(gTotalCtrl))
			{
				ar[gTotalCtrl++].ArcInitSession(out);
				out = new LinuxIoctlInterface();
			}
			else
			{
				unlink(buf);
			}
		}


		if (gTotalCtrl)
		{
			break;
		}
	}

	return gTotalCtrl;

}

#endif


#define ARECA_MIB 1,3,6,1,4,1,18928,1,1
int register_arecamib_sata(void)
{
	oid             areca_variables_oid[] = { 1, 3, 6, 1, 4, 1, 18928, 1, 1};
	oid             raidsys_oid[] = { ARECA_MIB, 1};
	oid             hwmon_oid[] = { ARECA_MIB, 2};
	oid             hdd_oid[] = { ARECA_MIB, 3, 1, 1};
	oid             raidset_oid[] = { ARECA_MIB, 4, 1, 1};
	oid             volset_oid[] = { ARECA_MIB, 5, 1, 1};
	oid             evt_oid[] = { ARECA_MIB, 6, 1, 1};
	oid             end_oid[] = { ARECA_MIB, 7, 1, 1};
	/*
	 * Register ourselves with the agent to handle our mib tree.
	 * The arguments are:
	 *    descr:   A short description of the mib group being loaded.
	 *    var:     The variable structure to load.
	 *                  (the name of the variable structure defined above)
	 *    vartype: The type of this variable structure
	 *    theoid:  The OID pointer this ARECA_MIB is being registered underneath.
	 */


	// register SATA controllers
	for (int i = 0; i < gTotalCtrl - gTotalSASCtrl; i++)
	{
		raidsys_oid[8] = 1 + i;
		hwmon_oid[8] = 1 + i;
		hdd_oid[8] = 1 + i;
		raidset_oid[8] = 1 + i;
		volset_oid[8] = 1 + i;
		evt_oid[8] = 1 + i;
		end_oid[8] = 1 + i;


		/*
		register_mib_context(const char *moduleName,
					 struct variable *var,
					 size_t varsize,
					 size_t numvars,
					 oid * mibloc,
					 size_t mibloclen,
					 int priority,
					 int range_subid,
					 oid range_ubound,
					 netsnmp_session * ss,
					 const char *context, int timeout, int flags)
					 */
		//REGISTER_MIB("raidsys_sata", raidsys_vars, variable1, raidsys_oid);
		register_mib_context("raidsys_sata", 
							 (struct variable *)raidsys_vars, 
							 sizeof(variable1),
							 sizeof(raidsys_vars) / sizeof(variable1), 
							 raidsys_oid,
							 sizeof(raidsys_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );

		//REGISTER_MIB("hwmon_sata", hwmon_vars, variable3, hwmon_oid);
		register_mib_context("hwmon_sata", 
							 (struct variable *)hwmon_vars, 
							 sizeof(variable3),
							 sizeof(hwmon_vars) / sizeof(variable3), 
							 hwmon_oid,
							 sizeof(hwmon_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );

		//REGISTER_MIB("hdd_sata", hdd_vars, variable1, hdd_oid);
		register_mib_context("hdd_sata", 
							 (struct variable *)hdd_vars, 
							 sizeof(variable1),
							 sizeof(hdd_vars) / sizeof(variable1), 
							 hdd_oid,
							 sizeof(hdd_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );

		//REGISTER_MIB("raidset_sata", raidset_vars, variable1, raidset_oid);
		register_mib_context("raidset_sata", 
							 (struct variable *)raidset_vars, 
							 sizeof(variable1),
							 sizeof(raidset_vars) / sizeof(variable1), 
							 raidset_oid,
							 sizeof(raidset_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );

		//REGISTER_MIB("volset_sata", volset_vars, variable1, volset_oid);		
		register_mib_context("volset_sata", 
							 (struct variable *)volset_vars, 
							 sizeof(variable1),
							 sizeof(volset_vars) / sizeof(variable1), 
							 volset_oid,
							 sizeof(volset_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );

		//REGISTER_MIB("evt_sata", evt_vars, variable1, evt_oid);
		register_mib_context("evt_sata", 
							 (struct variable *)evt_vars, 
							 sizeof(variable1),
							 sizeof(evt_vars) / sizeof(variable1), 
							 evt_oid,
							 sizeof(evt_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );
	}

	snmp_log(LOG_ERR,"SATA ARECA_MIB Registered OK.\n");

	return 1;
}



int register_arecamib_sas(void)
{
	oid             areca_variables_oid[] = { 1, 3, 6, 1, 4, 1, 18928, 1, 1};
	oid             raidsys_oid[] = { ARECA_MIB, 1};
	oid             hwmon_oid[] = { ARECA_MIB, 2, 1};
	oid             hdd_oid[] = { ARECA_MIB, 3, 1};
	oid             raidset_oid[] = { ARECA_MIB, 4, 1, 1};
	oid             volset_oid[] = { ARECA_MIB, 5, 1, 1};
	oid             evt_oid[] = { ARECA_MIB, 6, 1, 1};
	oid             end_oid[] = { ARECA_MIB, 7, 1, 1};


	// register SAS controllers
	for (int i = 0; i < gTotalSASCtrl; i++)
	{
		raidsys_oid[8] = 129 + i;
		hwmon_oid[8] = 129 + i;
		hdd_oid[8] = 129 + i;
		raidset_oid[8] = 129 + i;
		volset_oid[8] = 129 + i;
		evt_oid[8] = 129 + i;




		/*
		register_mib_context(const char *moduleName,
					 struct variable *var,
					 size_t varsize,
					 size_t numvars,
					 oid * mibloc,
					 size_t mibloclen,
					 int priority,
					 int range_subid,
					 oid range_ubound,
					 netsnmp_session * ss,
					 const char *context, int timeout, int flags)
					 */

		//REGISTER_MIB("raidsys_sas", raidsys_vars_sas, variable1, raidsys_oid);
		register_mib_context("raidsys_sas", 
							 (struct variable *)raidsys_vars_sas, 
							 sizeof(variable1),
							 sizeof(raidsys_vars_sas) / sizeof(variable1), 
							 raidsys_oid,
							 sizeof(raidsys_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );


		// register 9 enclosure hwmons
		for (int j = 0; j < 9; j++)
		{
			hwmon_oid[10] = j+1;            

			//REGISTER_MIB("hwmon_sas", hwmon_vars_sas, variable3, hwmon_oid);			
			register_mib_context("hwmon_sas", 
								 (struct variable *)hwmon_vars_sas, 
								 sizeof(variable3),
								 sizeof(hwmon_vars_sas) / sizeof(variable3), 
								 hwmon_oid,
								 sizeof(hwmon_oid)/sizeof(oid),
								 DEFAULT_MIB_PRIORITY,
								 0,
								 0,
								 NULL,
								 "", -1, 0 );
		}

		// register 8 enclosure hdds
		for (int j = 0; j < 8; j++)
		{
			hdd_oid[10] = j+1;
			//REGISTER_MIB("hdd_sas", hdd_vars_sas, variable3, hdd_oid);
			register_mib_context("hdd_sas", 
								 (struct variable *)hdd_vars_sas, 
								 sizeof(variable3),
								 sizeof(hdd_vars_sas) / sizeof(variable3), 
								 hdd_oid,
								 sizeof(hdd_oid)/sizeof(oid),
								 DEFAULT_MIB_PRIORITY,
								 0,
								 0,
								 NULL,
								 "", -1, 0 );
		}

		//REGISTER_MIB("raidset_sas", raidset_vars_sas, variable1, raidset_oid);
		register_mib_context("raidset_sas", 
							 (struct variable *)raidset_vars_sas, 
							 sizeof(variable1),
							 sizeof(raidset_vars_sas) / sizeof(variable1), 
							 raidset_oid,
							 sizeof(raidset_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );

		//REGISTER_MIB("volset_sas", volset_vars_sas, variable1, volset_oid);		
		register_mib_context("volset_sas", 
							 (struct variable *)volset_vars_sas, 
							 sizeof(variable1),
							 sizeof(volset_vars_sas) / sizeof(variable1), 
							 volset_oid,
							 sizeof(volset_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );

		//REGISTER_MIB("evt_sas", evt_vars_sas, variable1, evt_oid);
		register_mib_context("evt_sas", 
							 (struct variable *)evt_vars_sas, 
							 sizeof(variable1),
							 sizeof(evt_vars_sas) / sizeof(variable1), 
							 evt_oid,
							 sizeof(evt_oid)/sizeof(oid),
							 DEFAULT_MIB_PRIORITY,
							 0,
							 0,
							 NULL,
							 "", -1, 0 );
	}


	snmp_log(LOG_ERR,"SAS ARECA_MIB Registered OK.\n");
	return 1;
}



/*
 * This function is called at the time the agent starts up
 *  to do any initializations that might be required.
 *
 * In theory it is optional and can be omitted if no
 *  initialization is needed.  In practise, every module
 *  will need to register itself (or the objects being
 *  implemented will not appear in the ARECA_MIB tree), and this
 *  registration is typically done here.
 *
 * If this function is added or removed, you must re-run
 *  the configure script, to detect this change.
 */
void
init_arecamib(void)
{

	snmp_log(LOG_ERR,"init_areca(), init date:%s time:%s\n", __DATE__, __TIME__);

	DiscoveryDevice();

#if defined(__linux__)
	if (gTotalCtrl == 0)
	{
		DiscoveryCardDeviceThroughPassIoctl();
	}
#endif

	int i = 0, j = 0, t = 0;
	sSYSTEM_INFO sysinfo;

	memset(&sysinfo, 0, sizeof(sSYSTEM_INFO));

	while (t < gTotalCtrl)
	{
		if ( (ar[t].ArcGetSysInfo(&sysinfo) == ARC_SUCCESS) && (sysinfo.gsiTargetType == TARGET_SAS) )
		{
			arrSASRaidCard[i] = t;
			ar[t].setIdx(i);
			i++;
			gTotalSASCtrl = i;
		}
		else
		{
			arrRaidCard[j] = t;
			ar[t].setIdx(i);
			j++;
		}

		t++;
	}

	snmp_log(LOG_ERR,"[%d] devices found. [%d] for SAS & [%d] for Non-SAS\n", gTotalCtrl, gTotalSASCtrl, gTotalCtrl - gTotalSASCtrl);
	if ( gTotalCtrl == 0 )
	{
		return;
	}

	register_arecamib_sata();
	register_arecamib_sas();

	// ignore the old events
	int nCount = 0;
	for ( int index = 0; index < gTotalCtrl; index++ )
	{
		ar[index].ArcGetReqEventPage(FLAG_ALL, &nCount);
	}

	snmp_alarm_register(1, SA_REPEAT, EventMonitor, NULL);

	snmp_log(LOG_ERR,"snmpd_alarm_registered OK\n");

	register_config_handler("arcsnmp", "community", areca_parse_config_rocommunity, NULL, "raidctrl snmp community");
	register_config_handler("arcsnmp", "evtlevel", areca_parse_config_evtlevel, NULL, "raidctrl event notification level");

}

/*********************
*
*  Configuration file handling functions
*
*********************/

void
areca_parse_config_rocommunity(const char *token, char *cptr)
{
	/*
	 * Make sure the string fits in the space allocated for it.
	 */  

	int index = 0;
	char *ch = cptr;

	if ( cptr == NULL )
		return;

	if ( !strlen(cptr) )
		return;

	strcpy((char *)gCommunity, cptr);


	gCommunity[strlen(cptr)] = '\0';


	snmp_log(LOG_ERR,"rocommunity = %s\n", gCommunity);
}

void
areca_parse_config_rwcommunity(const char *token, char *cptr)
{
	/*
	 * Make sure the string fits in the space allocated for it.
	 */  

	int index = 0;
	char *ch = cptr;


	if ( cptr == NULL )
		return;

	if ( !strlen(cptr) )
		return;

	strcpy((char *)gCommunity, cptr);

	gCommunity[strlen(cptr)] = '\0';

	snmp_log(LOG_ERR,"rwcommunity = %s\n", gCommunity);
}

void
areca_parse_config_evtlevel(const char *token, char *cptr)
{
	if ( cptr != NULL )
	{
		gEvt_level = atoi(cptr);
	}

	snmp_log(LOG_ERR,"event level = %d\n", gEvt_level);
}


char *SnmpUtilIdsToA(oid *name, size_t *length, char *buf)
{

#if 1
	memset(buf, 0, OIDBUF);

	if ( name == NULL || length == NULL )
	{
		return NULL;
	}

	if ( (int)(*length) <= 0 )
	{
		return NULL;
	}

	sprint_mib_oid(buf, name, *length);
#endif
	return buf;
}



// 1.3.6.1.4.1.18928.1.1.1.1.0
// 1.3.6.1.4.1.18928.1.129.1.1.0
u_char *var_sys( struct variable *vp,
				 oid * name,
				 size_t * length,
				 int exact, 
				 size_t * var_len, 
				 WriteMethod ** write_method)
{
	ARC_STATUS      stat = ARC_SUCCESS;
	char            *result_str = NULL;
	int             result_len = 0; 
	char            *szCommunity = gCommunity;
	oid             ret_oid[20];
	unsigned char   *var = NULL;
	unsigned char   tag = 0;    
	int             ret_var_len = 0;    
	oid             newname[MAX_OID_LEN];
	sSYSTEM_INFO    sysinfo;

	*write_method  = NULL;

	stat = ar[gCurrentCtrl].ArcSNMPWithOid( (long *)name, // oid
											(int *)length, // oid length
											(unsigned char *)szCommunity, 
											(long *)ret_oid, // returned oid
											&tag, 
											&var, 
											(int *)&ret_var_len,
											(unsigned char)exact );
	if ( stat == ARC_SUCCESS )
	{
		if ( tag == ASN_INTEGER )
		{
			vp->type = ASN_INTEGER;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}           
			*var_len = ret_var_len;
			long_ret = *((long *)var);                                  
			return(u_char *)&long_ret;
		}
		else if (tag == ASN_OCTET_STR)
		{
			vp->type = ASN_OCTET_STR;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}
			result_str = (char *)var;
			result_len = ret_var_len;
			result_str[result_len] = 0;

			sprintf(string_ret,"%s", result_str);
			/*
			 * Note that the assumption that the answer will be an
			 *  integer does not hold true in this case, so the length
			 *  of the answer needs to be set explicitly.           
			 */
			*var_len = strlen(string_ret);          
			return(u_char *)string_ret;
		}
		else
		{
#ifdef DEBUG
			snmp_log(LOG_ERR,"SNMP ERROR\n");         
#endif
			return NULL;
		}  
	}
	else
	{
		snmp_log(LOG_ERR,"Failed to ArcSnmp, state = 0x%X\n", stat);        
	}


	return NULL;
}


// 1.3.6.1.4.1.18928.1.1.2
// 1.3.6.1.4.1.18928.1.129.2

u_char *var_hwmon( struct variable *vp,
				   oid * name,
				   size_t * length,
				   int exact, 
				   size_t * var_len, 
				   WriteMethod ** write_method)
{
	ARC_STATUS      stat = ARC_SUCCESS;
	char            *result_str = NULL;
	int             result_len = 0;
	char            *szCommunity = gCommunity;
	oid             ret_oid[20];
	unsigned char   *var = NULL;
	unsigned char   tag = 0;    
	int             ret_var_len = 0;
	oid             newname[MAX_OID_LEN];
	sSYSTEM_INFO    sysinfo;

	stat = ar[gCurrentCtrl].ArcSNMPWithOid( (long *)name, // oid
											(int *)length, // oid length
											(unsigned char *)szCommunity, 
											(long *)ret_oid, // returned oid
											&tag, 
											&var, 
											(int *)&ret_var_len,
											(unsigned char)exact );
	if ( stat == ARC_SUCCESS )
	{
		if ( tag == ASN_INTEGER )
		{
			vp->type = ASN_INTEGER;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}

			*var_len = ret_var_len;
			long_ret = *((long *)var);                                        
			return(u_char *)&long_ret;
		}
		else if (tag == ASN_OCTET_STR)
		{
			vp->type = ASN_OCTET_STR;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}
			result_str = (char *)var;
			result_len = ret_var_len;
			result_str[result_len] = 0;

			sprintf(string_ret,"%s", result_str);
			/*
			 * Note that the assumption that the answer will be an
			 *  integer does not hold true in this case, so the length
			 *  of the answer needs to be set explicitly.           
			 */
			*var_len = strlen(string_ret);                      
			return(u_char *)string_ret;
		}
		else
		{
#ifdef DEBUG
			snmp_log(LOG_ERR,"SNMP ERROR\n");         
#endif
			return NULL;
		}  
	}
	else
	{
		snmp_log(LOG_ERR,"Failed to ArcSnmp, state = 0x%X\n", stat);        
	}

	return NULL;
}


// 1.3.6.1.4.1.18928.1.1.3
// 1.3.6.1.4.1.18928.1.129.3

u_char *var_hdd( struct variable *vp,
				 oid * name,
				 size_t * length,
				 int exact, 
				 size_t * var_len, 
				 WriteMethod ** write_method)
{
	ARC_STATUS      stat = ARC_SUCCESS;
	char            *result_str = NULL;
	int             result_len = 0;
	char            *szCommunity = gCommunity;
	oid             ret_oid[20];
	unsigned char   *var = NULL;
	unsigned char   tag = 0;    
	int             ret_var_len = 0;
	oid             newname[MAX_OID_LEN];
	sSYSTEM_INFO    sysinfo;

	stat = ar[gCurrentCtrl].ArcSNMPWithOid( (long *)name, // oid
											(int *)length, // oid length
											(unsigned char *)szCommunity, 
											(long *)ret_oid, // returned oid
											&tag, 
											&var, 
											(int *)&ret_var_len,
											(unsigned char)exact );
	if ( stat == ARC_SUCCESS )
	{
		if ( tag == ASN_INTEGER )
		{
			vp->type = ASN_INTEGER;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}

			*var_len = ret_var_len;
			long_ret = *((long *)var);                                        
			return(u_char *)&long_ret;
		}
		else if (tag == ASN_OCTET_STR)
		{
			vp->type = ASN_OCTET_STR;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}
			result_str = (char *)var;
			result_len = ret_var_len;
			result_str[result_len] = 0;

			sprintf(string_ret,"%s", result_str);
			/*
			 * Note that the assumption that the answer will be an
			 *  integer does not hold true in this case, so the length
			 *  of the answer needs to be set explicitly.           
			 */
			*var_len = strlen(string_ret);                      
			return(u_char *)string_ret;
		}
		else
		{
#ifdef DEBUG
			snmp_log(LOG_ERR,"SNMP ERROR\n");         
#endif
			return NULL;
		}  
	}
	else
	{
		snmp_log(LOG_ERR,"Failed to ArcSnmp, state = 0x%X\n", stat);        
	}


	return NULL;
}


// 1.3.6.1.4.1.18928.1.1.4.1.1.1.X
// 1.3.6.1.4.1.18928.1.129.4.1.1.1.X

u_char *var_raidset( struct variable *vp,
					 oid * name,
					 size_t * length,
					 int exact, 
					 size_t * var_len, 
					 WriteMethod ** write_method)
{
	ARC_STATUS      stat = ARC_SUCCESS;
	char            *result_str = NULL;
	int             result_len = 0;
	char            *szCommunity = gCommunity;
	oid             ret_oid[20];
	unsigned char   *var = NULL;
	unsigned char   tag = 0;    
	int             ret_var_len = 0;
	oid             newname[MAX_OID_LEN];

#if 0

	sRAID_MAP       raidMap;

	memset(&raidMap, 0, sizeof(raidMap));   
	stat = ar[gCurrentCtrl].ArcHelpBuildRaidSetMap(&raidMap);
	if (stat != ARC_SUCCESS)
	{
		return NULL;
	}

	if (raidMap.total == 0)
	{
		return NULL;
	}

#endif

	stat = ar[gCurrentCtrl].ArcSNMPWithOid( (long *)name, // oid
											(int *)length, // oid length
											(unsigned char *)szCommunity, 
											(long *)ret_oid, // returned oid
											&tag, 
											&var, 
											(int *)&ret_var_len,
											(unsigned char)exact );
	if ( stat == ARC_SUCCESS )
	{
		if ( tag == ASN_INTEGER )
		{
			vp->type = ASN_INTEGER;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}

			*var_len = ret_var_len;
			long_ret = *((long *)var);                                        
			return(u_char *)&long_ret;
		}
		else if (tag == ASN_OCTET_STR)
		{
			vp->type = ASN_OCTET_STR;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}
			result_str = (char *)var;
			result_len = ret_var_len;
			result_str[result_len] = 0;

			sprintf(string_ret,"%s", result_str);
			/*
			 * Note that the assumption that the answer will be an
			 *  integer does not hold true in this case, so the length
			 *  of the answer needs to be set explicitly.           
			 */
			*var_len = strlen(string_ret);                      
			return(u_char *)string_ret;
		}
		else
		{
#ifdef DEBUG		
			snmp_log(LOG_ERR,"SNMP ERROR\n");         
#endif
			return NULL;
		}  
	}
	else
	{
		snmp_log(LOG_ERR,"Failed to ArcSnmp, state = 0x%X\n", stat);        
	}

	return NULL;
}


// 1.3.6.1.4.1.18928.1.1.5.1.1.1.X
// 1.3.6.1.4.1.18928.1.129.5.1.1.1.X
u_char *var_volset( struct variable *vp,
					oid * name,
					size_t * length,
					int exact, 
					size_t * var_len, 
					WriteMethod ** write_method)
{
	ARC_STATUS      stat = ARC_SUCCESS;
	char            *result_str = NULL;
	int             result_len = 0;
	char            *szCommunity = gCommunity;
	oid             ret_oid[20];
	unsigned char   *var = NULL;
	unsigned char   tag = 0;    
	int             ret_var_len = 0;
	oid             newname[MAX_OID_LEN];   


#if 0
	sVOL_MAP        volMap;

	memset(&volMap, 0, sizeof(volMap)); 
	stat = ar[gCurrentCtrl].ArcHelpBuildVolumeSetMap(&volMap);
	if (stat != ARC_SUCCESS)
	{
		return NULL;
	}

	if (volMap.total == 0)
	{
		return NULL;
	}

#endif

	stat = ar[gCurrentCtrl].ArcSNMPWithOid( (long *)name, // oid
											(int *)length, // oid length
											(unsigned char *)szCommunity, 
											(long *)ret_oid, // returned oid
											&tag, 
											&var, 
											(int *)&ret_var_len,
											(unsigned char)exact );
	if ( stat == ARC_SUCCESS )
	{
		if ( tag == ASN_INTEGER )
		{
			vp->type = ASN_INTEGER;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}
			// restore the original subsystem oid            
			*var_len = ret_var_len;
			long_ret = *((long *)var);                                        
			return(u_char *)&long_ret;
		}
		else if (tag == ASN_OCTET_STR)
		{
			vp->type = ASN_OCTET_STR;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}
			result_str = (char *)var;
			result_len = ret_var_len;
			result_str[result_len] = 0;

			sprintf(string_ret,"%s", result_str);
			/*
			 * Note that the assumption that the answer will be an
			 *  integer does not hold true in this case, so the length
			 *  of the answer needs to be set explicitly.           
			 */
			*var_len = strlen(string_ret);                      
			return(u_char *)string_ret;
		}
		else
		{
#ifdef DEBUG			
			snmp_log(LOG_ERR,"SNMP ERROR\n");         
#endif
			return NULL;
		}  
	}
	else
	{
		snmp_log(LOG_ERR,"Failed to ArcSnmp, state = 0x%X\n", stat);        
	}

	return NULL;
}



// 1.3.6.1.4.1.18928.1.1.6
// 1.3.6.1.4.1.18928.1.129.6
u_char *var_evt( struct variable *vp,
				 oid * name,
				 size_t * length,
				 int exact, 
				 size_t * var_len, 
				 WriteMethod ** write_method)
{
	ARC_STATUS      stat = ARC_SUCCESS;
	char            *result_str = NULL;
	int             result_len = 0;
	char            *szCommunity = gCommunity;
	oid             ret_oid[20];
	unsigned char   *var = NULL;
	unsigned char   tag = 0;    
	int             ret_var_len = 0;
	oid             newname[MAX_OID_LEN];
	sSYSTEM_INFO    sysinfo;
	int             result = 0;
	int             evtNumber = 0;

	memset(&sysinfo, 0, sizeof(sysinfo));
	stat = ar[gCurrentCtrl].ArcGetSysInfo(&sysinfo);
	if (stat != ARC_SUCCESS)
	{
		return NULL;
	}


	/*
	 1.3.6.1.4.1.18928.1.1.6.1.1.1.X [ INDEX ]
	 1.3.6.1.4.1.18928.1.1.6.1.1.2.X [ event string ]	
	*/
	memcpy((char *)newname, (char *)vp->name, (int)vp->namelen * sizeof(oid));
	result = snmp_oid_compare(name, *length, newname, vp->namelen);
	if (exact)
	{
		if ( *length != 14 )
			return NULL;
		if ( name[13] < 0 || name[13] >= sysinfo.gsiEvents )
			return NULL;
	}
	else
	{
		if (*length >= 14)
		{

			// vp->name 1.3.6.1.4.1.18928.1.1.1.1
			// name     18928.1.1.6.1.1.2.1
			result = snmp_oid_compare(newname, 13, name, 13);
			if ( result < 0 )
				evtNumber = 0;
			else if ( result == 0 )
			{
				evtNumber = name[13];
			}
			else
			{
				return NULL;
			}

			if ( name[12] == 2 && name[13] >= sysinfo.gsiEvents - 1 )
			{
				//name[13] = sysinfo.gsiEvents - 1;
				//name[13] = 1;
				//return NULL;				
			}
		}
		else
		{
			if ( snmp_oid_compare(newname, *length, name, *length) < 0 )
				return NULL;
			evtNumber = 0;
		}

		newname[13] = evtNumber;
		memcpy((char *)name, (char *)newname, 14 * sizeof(oid));
		*length = 14;
	}

	stat = ar[gCurrentCtrl].ArcSNMPWithOid( (long *)name, // oid
											(int *)length, // oid length
											(unsigned char *)szCommunity, 
											(long *)ret_oid, // returned oid
											&tag, 
											&var, 
											(int *)&ret_var_len,
											(unsigned char)exact );
	if ( stat == ARC_SUCCESS )
	{
		if ( tag == ASN_INTEGER )
		{
			vp->type = ASN_INTEGER;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}

			*var_len = ret_var_len;
			long_ret = *((long *)var);                                        
			return(u_char *)&long_ret;
		}
		else if (tag == ASN_OCTET_STR)
		{
			vp->type = ASN_OCTET_STR;
			for (int i=0; i<*length; i++)
			{
				name[i] = ret_oid[i];
			}
			result_str = (char *)var;
			result_len = ret_var_len;
			result_str[result_len] = 0;

			sprintf(string_ret,"%s", result_str);
			/*
			 * Note that the assumption that the answer will be an
			 *  integer does not hold true in this case, so the length
			 *  of the answer needs to be set explicitly.           
			 */
			*var_len = strlen(string_ret);                      
			return(u_char *)string_ret;
		}
		else
		{
#ifdef DEBUG
			snmp_log(LOG_ERR,"SNMP ERROR\n");         
#endif
			return NULL;
		}  
	}
	else
	{
		snmp_log(LOG_ERR,"Failed to ArcSnmp, state = 0x%X\n", stat);        
	}


	return NULL;


}


/*********************
 *
 *  System specific implementation functions
 *
 *********************/

/*
 * Define the callback function used in the areca_variables structure.
 * This is called whenever an incoming request refers to an object
 *  within this sub-tree.
 *
 * Four of the parameters are used to pass information in.
 * These are:
 *    vp      The entry from the 'areca_variables' array for the
 *             object being queried.
 *    name    The OID from the request.
 *    length  The length of this OID.
 *    exact   A flag to indicate whether this is an 'exact' request
 *             (GET/SET) or an 'inexact' one (GETNEXT/GETBULK).
 *
 * Four of the parameters are used to pass information back out.
 * These are:
 *    name     The OID being returned.
 *    length   The length of this OID.
 *    var_len  The length of the answer being returned.
 *    write_method   A pointer to the SET function for this object.
 *
 * Note that name & length serve a dual purpose in both roles.
 */

u_char *var_areca( struct variable *vp,
				   oid * name,
				   size_t * length,
				   int exact, 
				   size_t * var_len, 
				   WriteMethod ** write_method)
{
	/*
	 *  The result returned from this function needs to be a pointer to
	 *    static data (so that it can be accessed from outside).
	 *  Define suitable variables for any type of data we may return.
	 */

	ARC_STATUS      stat = ARC_SUCCESS;
	u_char          *result = NULL;     
	unsigned int    tmpOid;    
	char buf[OIDBUF];


	*write_method = NULL;

#if DEBUG
	snmp_log(LOG_ERR,"\n");
	snmp_log(LOG_ERR,"Entering var_areca()....\n");
#endif
	// the location 8 of oid number is the subsystem OID,
	// and it is the controller number too, we just save it
	// for subsequent operations.
	tmpOid = (unsigned int)name[8];

	if ((unsigned int)name[8] >= 0x80)
	{
		// SAS RaidCard
		gCurrentCtrl = (unsigned int)name[8] - 0x80 - 1;

		if (gCurrentCtrl >= gTotalSASCtrl)
		{
			snmp_log(LOG_ERR,"Invalid Ctrl(SAS) = %d\n", gCurrentCtrl);
			return NULL;
		}

		gCurrentCtrl = arrSASRaidCard[gCurrentCtrl];

		// since we just have 1 copy of mib table, so we must replace 
		// the original subsystem oid with "2" to look it up in the mib table,
		// if finished we must restore the oid number to the original one.
		name[8] = 2;
		vp->name[8] = 2;
#if DEBUG
		snmp_log(LOG_ERR,"Current Ctrl(SAS) = %d\n", gCurrentCtrl);
#endif
	}
	else
	{
		// Non-SAS RaidCard
		gCurrentCtrl = (unsigned int)name[8] - 1;

		if (gCurrentCtrl >= (gTotalCtrl - gTotalSASCtrl) )
		{
			snmp_log(LOG_ERR,"Invalid Ctrl(Non-SAS) = %d\n", gCurrentCtrl);
			return NULL;
		}

		gCurrentCtrl = arrRaidCard[gCurrentCtrl];
		// since we just have 1 copy of mib table, so we must replace 
		// the original subsystem oid with "1" to look it up in the mib table,
		// if finished we must restore the oid number to the original one.
		name[8] = 1;
		vp->name[8] = 1;
#if DEBUG
		snmp_log(LOG_ERR,"Current Ctrl(Non-SAS) = %d\n", gCurrentCtrl);
#endif
	}


#if DEBUG
	memset(buf, 0, sizeof(buf));    
	sprint_mib_oid(buf, vp->name, vp->namelen); 
	snmp_log(LOG_ERR,"  vp->name: %s vp->length: %d\n", buf , (int)vp->namelen);

	memset(buf, 0, sizeof(buf));
	sprint_mib_oid(buf, name, *length); 
	snmp_log(LOG_ERR," [in] name: %s length: %d exact: %d\n", buf , (int)*length, exact);   
#endif
	// 1.3.6.1.4.1.18928.1.1[2,3,4].1.1.0
	// 1.3.6.1.4.1.18928.1.129[130,131,132].1.1.0
	*write_method = NULL;
	*var_len = sizeof(LONG);

	if (name[9] == vp->name[9])
	{
		switch (name[9])
		{
		case 1:	// sys
			result = var_sys( vp, name, length, exact, var_len, write_method);      
			break;
		case 2:	// h/w mon
			result = var_hwmon( vp, name, length, exact, var_len, write_method);                
			break;
		case 3:	// hdd
			result = var_hdd( vp, name, length, exact, var_len, write_method);                  
			break;
		case 4:	// raidset
			result = var_raidset( vp, name, length, exact, var_len, write_method);                  
			break;
		case 5:	// volumeset
			result = var_volset( vp, name, length, exact, var_len, write_method);                   
			break;
		case 6:	// event
			result = var_evt( vp, name, length, exact, var_len, write_method);
			break;          
		default:
			break;
		}       
	}

#if DEBUG
	snmp_log(LOG_ERR,"------------------------------------------------------------\n");
	memset(buf, 0, sizeof(buf));    
	sprint_mib_oid(buf, name, *length); 
	snmp_log(LOG_ERR,"[out] name: %s length: %d exact: %d\n", buf , (int)*length, exact);

	if (result == NULL)
	{
		snmp_log(LOG_ERR,"[out] value = NULL\n");       
	}
	else
	{
		if (vp->magic == ARECASTRING)
		{
			snmp_log(LOG_ERR,"[out] var_len: %d\n[out] value: %s\n", *var_len, result); 
		}
		else if (vp->magic == ARECAINTEGER)
		{
			snmp_log(LOG_ERR,"[out] var_len: %d\n[out] value: %ld\n", *var_len, *((long *)result));
		}
		else
		{
			snmp_log(LOG_ERR,"Unknown tag\n");
		}
	}

	snmp_log(LOG_ERR,"Leaving var_areca()....\n");
#endif
	// restore the original subsystem oid
	name[8] = tmpOid;

	return result;
}

void 
EventMonitor( unsigned int clientreg,
			  void *clientarg)
{

	oid objid_enterprisetrap[] = { 1, 3, 6, 1, 4, 1, 18928, 1, 1};
	oid objid_rscreatelist[] = { 1, 3, 6, 1, 2, 1, 1, 1, 0};

	CArclibSNMP     *ctrl;
	ARC_STATUS      stat = ARC_SUCCESS;
	int             nCount = 0;				// total number of event objects we got	
	int             i;  
	int             nEventFlag = FLAG_NEW;	// the flag pass into ArcGetReqEventPage(...)
	int             nCurEvtLevel = EVT_SERIOUS;
	sSYSTEM_INFO    sysInfo;
	BYTE            ptr1[64];
	char            tmpStr[64];
	BOOL            gotEvent = FALSE;
	char            eventStrBuf[512];
	static  int CtrlIndex = 0;


	BYTE *eventWLevel[] ={
		(BYTE *)"URGENT",
		(BYTE *)"SERIOUS",
		(BYTE *)"WARNING",
		(BYTE *)"MESSAGE"
	};


	if (gTotalCtrl == 0)
		return;


	CtrlIndex = CtrlIndex % gTotalCtrl;

	ctrl = &ar[CtrlIndex];

	memset(ptr1, 0, sizeof(ptr1));
	memset(tmpStr, 0, sizeof(tmpStr));
	memset(eventStrBuf, 0, sizeof(eventStrBuf));
	// event object defined in event.h
	pEVENT_DATA pEventObject = (pEVENT_DATA)new sEVENT_DATA;    

	memset(pEventObject, 0, sizeof(sEVENT_DATA));

	stat = ctrl->ArcGetSysInfo(&sysInfo);
	if (stat != ARC_SUCCESS)
	{
		goto quit;
	}

	/* 
		We should call ArcSetTime() to correct the raid-subsystem time
	*/
	/* 
		Call ArcGetReqEventPage() to get the required event page from raid subsystem.
		There are 4 event pages reside in the raid subsystem and each contains 16 event objects.			
	*/
	stat = ctrl->ArcGetReqEventPage(nEventFlag, &nCount);
	if (stat == ARC_SUCCESS)
	{
		if (nCount != 0)
		{
			for ( i = nCount - 1; i >= 0; i--)
			{
				/*
					Call ArcGetEventObject function to extract the individual event object from event page
				*/
				if (ctrl->ArcGetEventObject( i, pEventObject ))
				{
					/* 
						Parse the event object we got
					*/
					switch ((int)pEventObject->evtCategory)
					{
					/* 
						The member "evtStr" of event object structure contains the string to indicate
						which raidset owns this event object
					*/
					case EVENT_RAIDSET:		// raidset event	
						nCurEvtLevel = raidEventLevel[pEventObject->evtType];
						strcpy( (char *)ptr1, htmRaidEvent[pEventObject->evtType]); 
						memcpy(tmpStr, pEventObject->evtStr, 16);
						sprintf((char *)eventStrBuf, "Controller#%d %s %s",CtrlIndex+1, tmpStr, ptr1);
						gotEvent = TRUE;
						break;                          
						/* 
							The member "evtStr" of event object structure contains the string to indicate
							which volumeset owns this event object
						*/
					case EVENT_VOLUMESET: // volumeset event
						nCurEvtLevel = volEventLevel[pEventObject->evtType];
						strcpy( (char *)ptr1, htmVolEvent[pEventObject->evtType]);  
						memcpy(tmpStr, pEventObject->evtStr, 16);
						sprintf((char *)eventStrBuf, "Controller#%d %s %s",CtrlIndex+1, tmpStr, ptr1);
						gotEvent = TRUE;
						break;

					case EVENT_DEVICE:	// device event
						nCurEvtLevel = devEventLevel[pEventObject->evtType];
						sprintf(tmpStr, "IDE Channel #%2d", pEventObject->evtChannel + 1);
						sprintf((char *)eventStrBuf, "Controller#%d %s %s",CtrlIndex+1, tmpStr, htmDevEvent[pEventObject->evtType] );
						gotEvent = TRUE;
						break;

					case EVENT_HOST:	// host event
						/* this is Host SCSI events */
						nCurEvtLevel = hostEventLevel[pEventObject->evtType];
						if ( sysInfo.gsiScsiHostChannels)
						{
							sprintf(tmpStr, "%s %s\n",
									eventCat[pEventObject->evtCategory], 
									htmScsiHostEvent[pEventObject->evtType]);
						}
						else
						{
							sprintf(tmpStr, "%s %s\n",
									eventCat[pEventObject->evtCategory], 
									htmIdeHostEvent[pEventObject->evtType]);
						}
						sprintf((char *)eventStrBuf, "Controller#%d %s",CtrlIndex+1, tmpStr);
						gotEvent = TRUE;
						break;

					case EVENT_HW_MONITOR:	// hardware monitor event
						nCurEvtLevel = hwEventLevel[pEventObject->evtType]; 
						memcpy(tmpStr, eventCat[pEventObject->evtCategory], 16);
						sprintf((char *)eventStrBuf, "Controller#%d %s %s",CtrlIndex+1, tmpStr, htmHwMonEvent[pEventObject->evtType] );
						gotEvent = TRUE;
						break;

					case EVENT_NEW_83782D:
						nCurEvtLevel = newEventLevel[pEventObject->evtType]; 
						memcpy(tmpStr, pEventObject->evtStr, 16);
						sprintf((char *)eventStrBuf, "Controller#%d %s %s",CtrlIndex+1, tmpStr, htmNewEventStr[pEventObject->evtType] );
						gotEvent = TRUE;
						break;
					case EVENT_NO_EVENT:
//							printf("		No Event\n");	
					default:
						break;  



					}// end switch
				}// end if

				if (gotEvent == TRUE)
				{
					if (gEvt_level != EVT_DISABLE && nCurEvtLevel <= gEvt_level)
					{
						netsnmp_variable_list * vars = NULL;

						int specificTrap = (int)pEventObject->evtCategory * 128 + (int)pEventObject->evtType;

						snmp_varlist_add_variable(&vars, objid_rscreatelist, OID_LENGTH(objid_rscreatelist),
												  ASN_OCTET_STR, (u_char *)eventStrBuf, strlen(eventStrBuf));


						// 
						if (sysInfo.gsiTargetType == TARGET_SAS)
						{
							objid_enterprisetrap[8] = ctrl->getIdx() + 129;
						}
						else
						{
							objid_enterprisetrap[8] = ctrl->getIdx() + 1;
						}
						send_enterprise_trap_vars(SNMP_TRAP_ENTERPRISESPECIFIC, specificTrap, objid_enterprisetrap,
												  OID_LENGTH(objid_enterprisetrap), vars);

						snmp_free_varbind(vars);

						snmp_log(LOG_ERR,"Ctrl[%d] issues the SNMP-Trap: %s\n", CtrlIndex, eventStrBuf);
					}
					// reset the goEvent flag
					gotEvent = FALSE;
				}
			}// end for
		}//end if
	}// end if

	quit:

	// monitor next controller
	CtrlIndex++;

	if (pEventObject)
		delete pEventObject;
}

#ifdef __cplusplus
}
#endif





